const std = @import("std");
const mem = std.mem;
const heap = std.heap;

pub const c = @cImport({
    @cInclude("libvirt/libvirt.h");
    @cInclude("libvirt/libvirt-admin.h");
    @cInclude("libvirt/libvirt-lxc.h");
    @cInclude("libvirt/libvirt-qemu.h");
    @cInclude("libvirt/virterror.h");
});

fn handleError() VirError {
    const err = c.virGetLastError();
    std.debug.print("err: {any}\n", .{err});
    return VirError.OK;
    // switch (err.*.code) {
    //     0 => return VirError.OK,
    //     1 => return VirError.InternalError,
    //     2 => return VirError.NoMemory,
    //     // TODO rest
    //     else => VirError.OK,
    // }
}

pub const VirError = error{
    OK,
    InternalError,
    NoMemory,
    NoSupport,
    UnknownHost,
    NoConnect,
    InvalidConn,
    InvalidDomain,
    InvalidArg,
    OperationFailed,
    GetFailed,
    PostFailed,
    HttpError,
    SExprError,
    NoXen,
    XenCall,
    OsType,
    NoKernel,
    NoRoot,
    NoSource,
    NoTarget,
    NoName,
    NoOs,
    NoDevice,
    NoXenstore,
    DriverFull,
    CallFailed,
    XmlError,
    DomExist,
    OperationDenied,
    OpenFailed,
    ReadFailed,
    ParseFailed,
    ConfSyntax,
    // TODO rest
};

fn string(str: [*c]const u8) []const u8 {
    return mem.span(str);
}

fn intFromFlags(comptime T: type, flags: []const T) c_uint {
    var flags_int: c_uint = 0;
    for (flags) |f| flags_int |= @intFromEnum(f);
    return flags_int;
}

fn Iterator(comptime T: type, comptime Ptr: type, comptime freeFn: *const fn (Ptr) callconv(.C) c_int) type {
    return struct {
        allocator: mem.Allocator,
        list: [*]Ptr,
        num: c_int,
        curr: usize,

        pub fn init(
            comptime PP: type, // Parent Pointer
            comptime P: type, // Parent
            comptime F: type, // Flags
            parent: *const P,
            flags: []const F,
            allocator: mem.Allocator,
            initFn: *const fn (?*PP, [*c][*c]Ptr, c_uint) callconv(.C) c_int,
        ) !Self {
            var list: [*]Ptr = undefined;
            const num = initFn(parent.ptr, @ptrCast(&list), intFromFlags(F, flags));
            return if (num < 0) handleError() else .{
                .allocator = allocator,
                .list = list,
                .num = num,
                .curr = 0,
            };
        }
        pub fn deinit(self: *Self) void {
            var i: usize = 0;
            while (i < self.num) : (i += 1) _ = freeFn(self.list[i]);
        }
        pub fn first(self: *Self) T {
            self.curr = 0;
            // return .{ .ptr = self.list[self.curr], .arena = heap.ArenaAllocator.init(self.allocator) };
            return T.init(self.list[self.curr], self.allocator);
        }
        pub fn next(self: *Self) ?T {
            if (self.curr >= self.num) return null;
            const ptr = self.list[self.curr];
            self.curr += 1;
            // return .{ .ptr = ptr, .arena = heap.ArenaAllocator.init(self.allocator) };
            return T.init(ptr, self.allocator);
        }
        const Self = @This();
    };
}

fn numOf(
    comptime P: type,
    ptr: P,
    fnPtr: *const fn (P) callconv(.C) c_int,
) !u32 {
    const num = fnPtr(ptr);
    return if (num < 0) handleError() else @intCast(num);
}

pub const Connection = struct {
    ptr: c.virConnectPtr,
    allocator: mem.Allocator,

    pub fn connect(uri: []const u8, allocator: mem.Allocator) VirError!Connection {
        const connection = c.virConnectOpenAuth(@ptrCast(uri), c.virConnectAuthPtrDefault, 0);
        if (connection) |conn| return .{
            .ptr = conn,
            .allocator = allocator,
        } else return handleError();
    }

    pub fn close(self: *const Connection) void {
        _ = c.virConnectClose(self.ptr);
    }

    pub fn getURI(self: *const Connection) ![]u8 {
        const uri = c.virConnectGetURI(self.ptr);
        defer std.c.free(uri);
        const str = string(uri);
        return if (str.len == 0) handleError() else try self.allocator.dupe(u8, str);
    }

    pub fn freeURI(self: *const Connection, uri: []u8) void {
        self.allocator.free(uri);
    }

    pub fn numOfPools(self: *const Connection) !u32 {
        return numOf(c.virConnectPtr, self.ptr, c.virConnectNumOfStoragePools);
    }
    pub fn numOfDefinedPools(self: *const Connection) !u32 {
        return numOf(c.virConnectPtr, self.ptr, c.virConnectNumOfDefinedStoragePools);
    }

    pub fn iteratePools(self: *const Connection, flags: []const Pool.ListFlags) !Pool.PoolIterator {
        return Pool.PoolIterator.init(
            c.virConnect,
            Connection,
            Pool.ListFlags,
            self,
            flags,
            self.allocator,
            c.virConnectListAllStoragePools,
        );
    }

    pub fn createPool(self: *const Connection, xml: []const u8, flags: []const Pool.CreateFlags) !Pool {
        const pool = c.virStoragePoolCreateXML(self.ptr, @ptrCast(xml), intFromFlags(Pool.CreateFlags, flags));
        return if (pool) |p| Pool.init(p, self.allocator) else handleError();
    }
    pub fn definePoolFlags(self: *const Connection, xml: []const u8, flags: []const Pool.DefineFlags) !Pool {
        const pool = c.virStoragePoolCreateXML(self.ptr, @ptrCast(xml), intFromFlags(Pool.DefineFlags, flags));
        return if (pool) |p| Pool.init(p, self.allocator) else handleError();
    }

    pub fn numOfDomains(self: *const Connection) !u32 {
        return numOf(c.virConnectPtr, self.ptr, c.virConnectNumOfDomains);
    }
    pub fn numOfDefinedDomains(self: *const Connection) !u32 {
        return numOf(c.virConnectPtr, self.ptr, c.virConnectNumOfDefinedDomains);
    }

    pub fn iterateDomains(self: *const Connection, flags: []const Domain.ListFlags) !Domain.DomainIterator {
        return Domain.DomainIterator.init(
            c.virConnect,
            Connection,
            Domain.ListFlags,
            self,
            flags,
            self.allocator,
            c.virConnectListAllDomains,
        );
    }

    pub fn createDomain(self: *const Connection, xml: []const u8, flags: []const Domain.CreateFlags) !Domain {
        const domain = c.virDomainCreateXML(self.ptr, @ptrCast(xml), intFromFlags(Domain.CreateFlags, flags));
        return if (domain) |d| Domain.init(d, self.allocator) else handleError();
    }
    pub fn defineDomain(self: *const Connection, xml: []const u8) !Domain {
        const domain = c.virDomainDefineXML(self.ptr, @ptrCast(xml));
        return if (domain) |d| Domain.init(d, self.allocator) else handleError();
    }
    pub fn defineDomainFlags(self: *const Connection, xml: []const u8, flags: []const Domain.DefineFlags) !Domain {
        const domain = c.virDomainDefineXMLFlags(self.ptr, @ptrCast(xml), intFromFlags(Domain.DefineFlags, flags));
        return if (domain) |d| Domain.init(d, self.allocator) else handleError();
    }
};

pub const Pool = struct {
    ptr: c.virStoragePoolPtr,
    arena: *heap.ArenaAllocator,

    pub fn init(ptr: c.virStoragePoolPtr, allocator: mem.Allocator) Pool {
        var arena = heap.ArenaAllocator.init(allocator);
        return .{
            .ptr = ptr,
            .arena = &arena,
        };
    }
    pub fn deinit(self: *const Pool) void {
        self.arena.deinit();
    }

    pub fn getName(self: *const Pool) ![]const u8 {
        const name = c.virStoragePoolGetName(self.ptr);
        // defer std.c.free(name);
        const str = string(name);
        return if (str.len == 0) handleError() else str;
        // return if (str.len == 0) handleError() else try self.arena.allocator().dupe(u8, str);
    }

    pub fn createVolume(pool: *const Pool, xml: []const u8, flags: []Volume.Flags) !Volume {
        const volume = c.virStorageVolCreateXML(pool, xml, intFromFlags(Volume.CreateFlags, flags));
        return .{ .ptr = volume };
    }

    pub fn iterateVolumes(self: *const Pool, flags: []const ListFlags) Volume.VolumeIterator {
        return Volume.VolumeIterator.init(
            c.virStorageVolumePtr,
            Pool,
            Volume.ListFlags,
            self,
            flags,
            self.arena.allocator(),
            c.virStoragePoolListAllVolumes,
        );
    }

    pub const ListFlags = enum(c_uint) {
        Inactive = c.VIR_CONNECT_LIST_STORAGE_POOLS_INACTIVE,
        Active = c.VIR_CONNECT_LIST_STORAGE_POOLS_ACTIVE,
        Persistent = c.VIR_CONNECT_LIST_STORAGE_POOLS_PERSISTENT,
        Transient = c.VIR_CONNECT_LIST_STORAGE_POOLS_TRANSIENT,
        Autostart = c.VIR_CONNECT_LIST_STORAGE_POOLS_AUTOSTART,
        NoAutostart = c.VIR_CONNECT_LIST_STORAGE_POOLS_NO_AUTOSTART,
        Dir = c.VIR_CONNECT_LIST_STORAGE_POOLS_DIR,
        Fs = c.VIR_CONNECT_LIST_STORAGE_POOLS_FS,
        NetFs = c.VIR_CONNECT_LIST_STORAGE_POOLS_NETFS,
        Logical = c.VIR_CONNECT_LIST_STORAGE_POOLS_LOGICAL,
        Disk = c.VIR_CONNECT_LIST_STORAGE_POOLS_DISK,
        Iscsi = c.VIR_CONNECT_LIST_STORAGE_POOLS_ISCSI,
        Scsi = c.VIR_CONNECT_LIST_STORAGE_POOLS_SCSI,
        Mpath = c.VIR_CONNECT_LIST_STORAGE_POOLS_MPATH,
        Rbd = c.VIR_CONNECT_LIST_STORAGE_POOLS_RBD,
        Sheepdog = c.VIR_CONNECT_LIST_STORAGE_POOLS_SHEEPDOG,
        Gluster = c.VIR_CONNECT_LIST_STORAGE_POOLS_GLUSTER,
        Zfs = c.VIR_CONNECT_LIST_STORAGE_POOLS_ZFS,
        Vstorage = c.VIR_CONNECT_LIST_STORAGE_POOLS_VSTORAGE,
        IscsiDirect = c.VIR_CONNECT_LIST_STORAGE_POOLS_ISCSI_DIRECT,
    };
    pub const BuildFlags = enum(c_uint) {};
    pub const CreateFlags = enum(c_uint) {};
    pub const DefineFlags = enum(c_uint) {};
    pub const DeleteFlags = enum(c_uint) {};

    pub const Info = struct {
        state: State,
        capacity: u64,
        allocation: u64,
        available: u64,
    };

    pub const State = enum(c_uint) {
        Inactive = c.VIR_STORAGE_POOL_INACTIVE,
        Building = c.VIR_STORAGE_POOL_BUILDING,
        Running = c.VIR_STORAGE_POOL_RUNNING,
        Degraded = c.VIR_STORAGE_POOL_DEGRADED,
        Inaccessible = c.VIR_STORAGE_POOL_INACCESSIBLE,
        Last = c.VIR_STORAGE_POOL_STATE_LAST,
    };
    pub const EventID = enum(c_uint) {
        Lifecycle = c.VIR_STORAGE_POOL_EVENT_ID_LIFECYCLE,
        Refresh = c.VIR_STORAGE_POOL_EVENT_ID_REFRESH,
        Last = c.VIR_STORAGE_POOL_EVENT_ID_LAST,
    };
    pub const EventLifecycleType = enum(c_uint) {
        Defined = c.VIR_STORAGE_POOL_EVENT_DEFINED,
        Undefined = c.VIR_STORAGE_POOL_EVENT_UNDEFINED,
        Started = c.VIR_STORAGE_POOL_EVENT_STARTED,
        Stopped = c.VIR_STORAGE_POOL_EVENT_STOPPED,
        Created = c.VIR_STORAGE_POOL_EVENT_CREATED,
        Deleted = c.VIR_STORAGE_POOL_EVENT_DELETED,
        Last = c.VIR_STORAGE_POOL_EVENT_LAST,
    };

    pub const PoolIterator = Iterator(Pool, c.virStoragePoolPtr, c.virStoragePoolFree);

    pub const Volume = struct {
        ptr: c.virStorageVolPtr,
        arena: *heap.ArenaAllocator,

        pub fn getName(self: *const Volume) ![]const u8 {
            const name = c.virStorageVolGetName(self.ptr);
            // defer std.c.free(name);
            const str = string(name);
            return if (str.len == 0) handleError() else str;
        }

        pub const ListFlags = enum(c_uint) {};
        pub const CreateFlags = enum(c_uint) {
            PreallocMetadata = c.VIR_STORAGE_VOL_CREATE_PREALLOC_METADATA,
            Reflink = c.VIR_STORAGE_VOL_CREATE_REFLINK,
            Validate = c.VIR_STORAGE_VOL_CREATE_VALIDATE,
        };
        pub const DeleteFlags = enum(c_uint) {
            Normal = c.VIR_STORAGE_VOL_DELETE_NORMAL,
            Zeroed = c.VIR_STORAGE_VOL_DELETE_ZEROED,
            WithSnapshots = c.VIR_STORAGE_VOL_DELETE_WITH_SNAPSHOTS,
        };
        pub const DownloadFlags = enum(c_uint) {
            SparseStream = c.VIR_STORAGE_VOL_DOWNLOAD_SPARSE_STREAM,
        };
        pub const InfoFlags = enum(c_uint) {
            UseAllocation = c.VIR_STORAGE_VOL_USE_ALLOCATION,
            GetPhysical = c.VIR_STORAGE_VOL_GET_PHYSICAL,
        };
        pub const ResizeFlags = enum(c_uint) {
            Allocate = c.VIR_STORAGE_VOL_RESIZE_ALLOCATE,
            Delta = c.VIR_STORAGE_VOL_RESIZE_DELTA,
            Shrink = c.VIR_STORAGE_VOL_RESIZE_SHRINK,
        };
        pub const UploadFlags = enum(c_uint) {};
        pub const Info = struct {
            type: Type,
            capacity: u64,
            allocation: u64,
        };
        pub const Type = enum(c_uint) {
            File = c.VIR_STORAGE_VOL_FILE,
            Block = c.VIR_STORAGE_VOL_BLOCK,
            Dir = c.VIR_STORAGE_VOL_DIR,
            Network = c.VIR_STORAGE_VOL_NETWORK,
            Netdir = c.VIR_STORAGE_VOL_NETDIR,
            Ploop = c.VIR_STORAGE_VOL_PLOOP,
            Last = c.VIR_STORAGE_VOL_LAST,
        };
        pub const WipeAlgorithm = enum(c_uint) {
            Zero = c.VIR_STORAGE_VOL_WIPE_ALG_ZERO,
            Nnsa = c.VIR_STORAGE_VOL_WIPE_ALG_NNSA,
            Dod = c.VIR_STORAGE_VOL_WIPE_ALG_DOD,
            Bsi = c.VIR_STORAGE_VOL_WIPE_ALG_BSI,
            Gutmann = c.VIR_STORAGE_VOL_WIPE_ALG_GUTMANN,
            Schneier = c.VIR_STORAGE_VOL_WIPE_ALG_SHNEIER,
            Pfitzner7 = c.VIR_STORAGE_VOL_WIPE_ALG_PFITZNER7,
            Pfitzner33 = c.VIR_STORAGE_VOL_WIPE_ALG_PFITZNER33,
            Random = c.VIR_STORAGE_VOL_WIPE_ALG_RANDOM,
            Trim = c.VIR_STORAGE_VOL_WIPE_ALG_TRIM,
            Last = c.VIR_STORAGE_VOL_WIPE_ALG_LAST,
        };

        pub const VolumeIterator = Iterator(Volume, c.virStorageVolPtr, c.virStorageVolFree);
    };
};

pub const Domain = struct {
    ptr: c.virDomainPtr,
    arena: *heap.ArenaAllocator,

    pub fn init(ptr: c.virDomainPtr, allocator: mem.Allocator) Domain {
        var arena = heap.ArenaAllocator.init(allocator);
        return .{
            .ptr = ptr,
            .arena = &arena,
        };
    }
    pub fn deinit(self: *const Domain) void {
        self.arena.deinit();
    }

    pub fn getName(self: *const Domain) []const u8 {
        const name = c.virDomainGetName(self.ptr);
        return mem.span(name);
    }

    pub fn isActive(self: *const Domain) bool {
        const active = c.virDomainIsActive(self.ptr);
        return if (active == 0) false else true;
    }

    pub const ListFlags = enum(c_uint) {
        Active = c.VIR_CONNECT_LIST_DOMAINS_ACTIVE,
        Inactive = c.VIR_CONNECT_LIST_DOMAINS_INACTIVE,
    };
    pub const AbortJobFlagsValues = enum(c_uint) {};
    pub const AgentResponseTimeoutValues = enum(c_uint) {};
    pub const AuthorizedSSHKeysSetFlags = enum(c_uint) {};
    pub const BackupBeginFlags = enum(c_uint) {};
    pub const BlockCommitFlags = enum(c_uint) {};
    pub const BlockCopyFlags = enum(c_uint) {};
    pub const BlockJobAbortFlags = enum(c_uint) {};
    pub const BlockJobInfoFlags = enum(c_uint) {};
    pub const BlockJobSetSpeedFlags = enum(c_uint) {};
    pub const BlockPullFlags = enum(c_uint) {};
    pub const BlockRebaseFlags = enum(c_uint) {};
    pub const BlockResizeFlags = enum(c_uint) {};
    pub const ChannelFlags = enum(c_uint) {};
    pub const ConsoleFlags = enum(c_uint) {};
    pub const CoreDumpFlags = enum(c_uint) {};
    pub const CreateFlags = enum(c_uint) {
        None = c.VIR_DOMAIN_NONE,
        StartPaused = c.VIR_DOMAIN_START_PAUSED,
        StartAutodestroy = c.VIR_DOMAIN_START_AUTODESTROY,
        StartBypassCache = c.VIR_DOMAIN_START_BYPASS_CACHE,
        StartForceBoot = c.VIR_DOMAIN_START_FORCE_BOOT,
        StartValidate = c.VIR_DOMAIN_START_VALIDATE,
        StartResetNvram = c.VIR_DOMAIN_START_RESET_NVRAM,
    };
    pub const DefineFlags = enum(c_uint) {
        Validate = c.VIR_DOMAIN_DEFINE_VALIDATE,
    };
    // TODO rest

    pub const BlockInfo = struct {};
    pub const BlockJobCursor = u64;
    pub const BlockJobInfo = struct {};
    pub const BlockStats = struct {};
    pub const ControlInfo = struct {};
    // TODO rest

    pub const BlockJobType = enum(c_uint) {};
    pub const BlockedReason = enum(c_uint) {};
    pub const ControlErrorReason = enum(c_uint) {};
    pub const ControlState = enum(c_uint) {};
    pub const CoreDumpFormat = enum(c_uint) {};
    pub const CrashedReason = enum(c_uint) {};
    // TODO rest

    pub const DomainIterator = Iterator(Domain, c.virDomainPtr, c.virDomainFree);
};
